# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import time
import os
import getpass
import signal
import sys
import platform
import subprocess
import time
import threading
import logging
from collections import OrderedDict

import config
import log
import rights
import paths
import platforms
import version
from plugin_controller import PluginController
from user_login import UserLogin
from updater import Updater
from rights import RightsChecker
from camera_controller import CameraController
from tray_controller import TrayController
from makerware_utils import ConveyorKillWaiter
from usb_detect import USBDetector
from network_detect import NetworkDetector
from client_detector import ClientDetector
from printer_interface import PrinterInterface
from gpio_interface import GPIOInterface
from detection_wizard import DetectionWizard


class App:

    LOG_TIMESTAMP_PERIOD = 30
    BROWSER_OPENING_DELAY = 2
    QUIT_THREAD_JOIN_TIMEOUT = 21
    LOOP_SLEEP_STEPS = 10

    def __init__(self):
        self.logger = log.create_logger(None)
        self.logger.info('Starting 3DPrinterOS client...')
        self.logger.info('Version: ' + version.full_version_string())
        self.logger.info('Operating system: ' + platform.system() + ' ' + platform.release())
        self.logger.info('Python: %d.%d.%d' % sys.version_info[0:3])
        if config.get_settings()['protocol']['encryption']:
            port_key_name = 'https_port'
        else:
            port_key_name = 'http_port'
        port = config.get_settings()['protocol'].get(port_key_name, '')
        if port:
            port = ":" + str(port)
        self.logger.info('Server: %s%s' % (config.get_settings()['URL'], port))
        try:
            self.init()
        except:
            self.logger.exception("Exception in App.init:")
        finally:
            self.quit()

    @log.log_exception
    def init(self):
        config.Config.instance().set_app_pointer(self)
        rights.add_cacert_to_certifi()
        self.printer_interfaces = []
        self.offline_mode = False
        self.new_printers_detection_wizard = False
        self.poweroff_flag = False
        self.virtual_printer_enabled = False
        self.stop_flag = False
        self.restart_flag = False
        self.invalid_usb_infos = []
        self.closing_status = []
        self.os_user_name = getpass.getuser()
        self.min_loop_time = config.get_settings()['main_loop_period']
        signal.signal(signal.SIGINT, self.intercept_signal)
        signal.signal(signal.SIGTERM, self.intercept_signal)
        self.plugin_controller = PluginController(self)
        self.rights_checker = RightsChecker(self)
        self.conveyor_kill_waiter = ConveyorKillWaiter(self)
        self.user_login = UserLogin(self, retry_in_background=config.get_settings()['pre_login_ui'])
        self.detection_wizard = DetectionWizard(self)
        self.init_user_interfaces()
        self.wait_and_open_browser()
        self.client_detector = ClientDetector(self)
        if self.user_login.wait() and self.rights_checker.wait() and self.conveyor_kill_waiter.wait():
            self.local_ip = self.user_login.http_connection.local_ip
            self.mac = self.user_login.http_connection.mac
            config.Config.instance().set_profiles(self.user_login.profiles)
            self.virtual_printer_enabled = config.get_settings()['virtual_printer']['enabled']
            self.virtual_printer_usb_info = dict(config.get_settings()['virtual_printer'])
            del self.virtual_printer_usb_info['enabled']
            self.camera_controller = CameraController(self)
            self.tray_controller = TrayController()
            self.usb_detector = USBDetector()
            self.network_printers_detector = NetworkDetector(self)
            self.updater = Updater(self)
            if not self.offline_mode:
                self.updater.start_checks_loop()
            self.main_loop()

    @staticmethod
    def open_browser(start=False):
        import logging 
        logger = logging.getLogger()
        if start and not config.get_settings()['web_interface']['browser_opening_on_start']:
            logger.info("Browser opening prevented by settings")
            return 
        address = "http://%s:%d" % (config.LOCAL_IP, config.get_settings()['web_interface']['port'])
        logger.info("Opening browser tab %s" % address)
        import webbrowser
        webbrowser.open(address, 2)

    def wait_and_open_browser(self):
        if config.get_settings()['web_interface']['browser_opening_on_start']:
            self.logger.info("Waiting for timeout or login to open browser...")
            start = time.monotonic()
            while time.monotonic() < start + self.BROWSER_OPENING_DELAY and not self.user_login.user_token:
                time.sleep(0.01)
            self.logger.info("...done")
            self.open_browser()

    def init_user_interfaces(self):
        self.gpio_interface = GPIOInterface()
        if config.get_settings()['web_interface']['enabled']:
            import web_interface 
            self.web_interface = self.init_user_interface(web_interface.WebInterface)
        if config.get_settings()['qt_interface']['enabled']:
            try:
                import qt_interface
            except ImportError:
                self.logger.error("Error on import of qt_interface - disabling it")
            else:
                self.qt_interface = self.init_user_interface(qt_interface.QtInterface)

    def init_user_interface(self, interface_class):
        interface = interface_class(self)
        interface.start()
        self.logger.debug("Waiting for %s to start..." % str(interface))
        interface.wait()
        self.logger.debug("...interface %s is up and running." % interface_class.__class__.__name__)
        return interface

    def intercept_signal(self, signal_code, frame):
        self.logger.warning("SIGINT or SIGTERM received. Starting exit sequence")
        if self.stop_flag:
            self.quit()
        self.stop_flag = True

    def toggle_virtual_printer(self):
        self.virtual_printer_enabled = not self.virtual_printer_enabled
        settings = config.get_settings()
        settings['virtual_printer']['enabled'] = self.virtual_printer_enabled
        config.Config.instance().save_settings(settings)

    @log.log_exception
    def main_loop(self):
        last_timestamp_time = 0
        while not self.stop_flag:
            loop_start_time = time.monotonic()
            detected_printers = []
            if not self.detection_wizard.active:
                detected_printers = self.usb_detector.get_printers_list()
                if self.virtual_printer_enabled:
                    detected_printers.append(self.virtual_printer_usb_info)
                detected_printers.extend(self.network_printers_detector.get_printers_list())
                self.connect_new_printers(detected_printers)
            for pi in self.printer_interfaces:
                if pi.usb_info not in detected_printers:
                    if not pi.forced_state == "error":
                        pi.register_error(99, "Printer is no longer detected as device", is_blocking=True)
                if not pi.is_alive():
                    pi.close()
                    self.logger.info('Removing %s from printer list' % pi.usb_info)
                    self.printer_interfaces.remove(pi)
            self.gpio_interface.check_events(self.printer_interfaces)
            loop_end_time = time.monotonic()
            if loop_end_time > last_timestamp_time + self.LOG_TIMESTAMP_PERIOD:
                self.logger.debug(time.strftime("%d %b %Y %H:%M:%S", time.localtime()))
                last_timestamp_time = loop_end_time
            sleep_time = loop_end_time - loop_start_time + self.min_loop_time
            if sleep_time > 0:
                steps_left = self.LOOP_SLEEP_STEPS
                while steps_left and not self.stop_flag:
                    time.sleep(sleep_time/self.LOOP_SLEEP_STEPS)
                    steps_left -= 1

    def validate_usb_info(self, usb_info):
        if isinstance(usb_info, dict):
            vid = usb_info.get('VID')
            pid = usb_info.get('VID')
            if pid and vid and isinstance(vid, str) and isinstance(pid, str):
                return True
        if not usb_info in self.invalid_usb_infos:
            self.invalid_usb_infos.append(usb_info)
            self.logger.error("Invalid printer info: " + str(usb_info))
        return False

    def connect_new_printers(self, detected_printers):
        currently_connected_usb_info = [pi.usb_info for pi in self.printer_interfaces]
        for usb_info in detected_printers:
            if usb_info not in currently_connected_usb_info:
                if self.validate_usb_info(usb_info):
                    pi = PrinterInterface(self, usb_info, config.get_settings()['printer_loop_period'])
                    pi.start()
                    currently_connected_usb_info.append(pi.usb_info)
                    self.printer_interfaces.append(pi)

    def close_module(self, state, closing_function, *args):
        self.logger.info(state)
        self.closing_status.append(state)
        closing_function(*args)
        self.closing_status[-1] += 'done'
        time.sleep(0.1)

    def register_error(self, code, message, is_blocking=False, is_info=False):
        self.logger.warning("Error:N%d in child of app. %s" % (code, message))

    def raise_stop_flag(self):
        self.stop_flag = True

    def set_offline_mode(self, enabled=True):
        self.offline_mode = enabled
        for pi in self.printer_interfaces:
            pi.offline_mode = enabled
        if enabled:
            self.user_login.init_profiles()

    @log.log_exception
    def quit(self): # TODO refactor all the closing by creating class Closable and subclasses for each case
        self.logger.info("Starting exit sequence...")
        if hasattr(self, 'client_detector'):
            self.close_module('Closing client detector...', self.client_detector.close)
        if hasattr(self, 'camera_controller'):
            self.close_module('Closing camera...', self.camera_controller.stop_camera_process)
        if hasattr(self, 'gpio_interface'):
            self.close_module('Closing GPIO interface...', self.gpio_interface.close)
        printer_interfaces = getattr(self, "printer_interfaces", [])
        for pi in printer_interfaces:
            printer_name = pi.printer_profile.get('name', 'nameless printer')
            state = 'Closing ' + printer_name + '...'
            self.close_module(state, pi.close)
        for pi in printer_interfaces:
            printer_name = pi.printer_profile.get('name', 'nameless printer')
            state = 'Joining ' + printer_name + '...'
            self.close_module(state, pi.join, self.QUIT_THREAD_JOIN_TIMEOUT)
        if hasattr(self, "plugin_controller"):
            self.close_module('Closing Plugins...', self.plugin_controller.close)
        if hasattr(self, 'tray_controller'):
            self.close_module('Closing tray subprocess...', self.tray_controller.close)
        time.sleep(0.2)  # need to reduce logging spam
        if hasattr(self, "web_interface"):
            self.close_module("Closing web server...", self.web_interface.close)
        if hasattr(self, "qt_interface"):
            self.close_module("Closing Qt interface...", self.qt_interface.close)
        # go through all existing threads and try to close lost ones
        for thread in threading.enumerate():
            if thread.is_alive:
                close_method = getattr(thread, "close", None)
                if close_method:
                    close_method()
                else:
                    thread.stop_flag = True
        self.logger.info("...exit sequence finish.")
        self.logger.info("Goodbye ;-)")
        if getattr(self, "poweroff_flag", False):
            if platforms.get_platform() == 'rpi' and rights.is_sudo_available():
                subprocess.call('sudo bash -c "echo 1 > /sys/class/backlight/rpi_backlight/bl_power"', shell=True)
                subprocess.call(["sudo", "poweroff"])
                time.sleep(1)
