# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import base64
import collections
import io
import logging
import os
import re
import time
import threading

import config
import filename
import platforms


class BaseSender:

    #TODO refactor gcodes, preprocess_gcodes, load_gcodes, unbuffered_gcodes mess
    #TODO implement print from disk(rolling buffer)

    TEMPERATURE_SIGNS_AFTER_DOT = 2 # rounding done using this
    REPORT_JOBS = False
    BUFFER_CLASS = list
    MEMORY_STORE_COOF = 2.5
    MEMORY_STORE_MARGIN = 60 * 1024 * 1024 # 60MB
    GCODES_PREPROCESS_BUFFER = 20 * 1024 * 1024 # 20MB

    def __init__(self, parent=None, usb_info={}, profile={}):
        if parent:
            self.logger = parent.logger.getChild(self.__class__.__name__)
        else:
            self.logger = logging.getLogger(self.__class__.__name__)
        self.stop_flag = False
        if parent: 
            self.parent = parent
        else:
            self.parent = FakeParent()
        self.usb_info = usb_info
        if not hasattr(self, "temps"):
            self.temps = [0.0, 0.0]
            self.target_temps = [0.0, 0.0]
        self.position = [0.0, 0.0, 0.0, 0.0]  # X, Y, Z, E
        self.profile = profile # empty profile is ok for some senders
        self.position = []  # [X, Y, Z, E]
        self.total_gcodes = 0
        self.estimated_time = None
        self.buffer = collections.deque()
        self.current_line_number = 0
        self.percent = 0
        self.operational_flag = False
        self.printing_flag = False
        self.pause_flag = False
        self.heating = False #TODO rename to heating_flag after merge
        self.callbacks_lock = threading.Lock()
        self.response_callbacks = [] # functions list to call on printer_response. if not empty it will suppress tempratures requesting.
        self.responses_planned = 0
        self.responses = []
        self.full_filename = ""
        self.displayed_filename = ""
        self.est_print_time = 0 # in seconds
        self.print_time_left = 0 # in seconds
        self.average_printing_speed = 0 # percents per second
        self.clouds_job_id = None
        self.printers_job_id = None
        self.is_base64_re = re.compile(b"^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{4}|[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)$")
        if not usb_info or not isinstance(usb_info, dict):
            raise RuntimeError("Invalid or empty printer id:" % usb_info)
        self.override_clouds_estimations = config.get_settings().get('print_estimation', {}).get('override_clouds', False)
        self.allow_increase_of_print_time_left = config.get_settings().get('print_estimation', {}).get('allow_rise_time_left', False)
        self.intercept_pause = config.get_settings().get('intercept_pause')
        #self.heating_start_time = None
        #self.last_pause_time = None
        #self.print_start_time = 0.0
        #self.sum_pause_and_heating_duration = 0 # in seconds
        #self.pause_and_heating_duration_lock = threading.Lock()
        self.is_base64_re = re.compile(b"^([A-Za-z0-9+/]{4})*([A-Za-z0-9+/]{4}|[A-Za-z0-9+/]{3}=|[A-Za-z0-9+/]{2}==)$")

    def set_total_gcodes(self, length):
        self.total_gcodes = length

    def get_total_gcodes(self):
        return self.total_gcodes

    def load_gcodes(self, gcodes):
        raise NotImplementedError

    def unbuffered_gcodes(self, gcodes):
        raise NotImplementedError

    def cancel(self):
        self.parent.register_error(605, "Cancel is not supported for this printer type", is_blocking=False)
        return False

    def preprocess_gcodes(self, gcodes_in):
        gcodes_out = self.BUFFER_CLASS()
        if gcodes_in:
            if type(gcodes_in) in (list, tuple, collections.deque):
                if type(gcodes_in[0]) == str:
                    sep = "\n"
                else:
                    sep = b"\b"
                gcodes_in = sep.join(gcodes_in)
            if type(gcodes_in) == str:
                partline = ""
            else:
                partline = b""
            while gcodes_in:
                if self.parent and self.parent.stop_flag:
                    break 
                buf_len = min(len(gcodes_in), self.GCODES_PREPROCESS_BUFFER)
                buf = partline + gcodes_in[:buf_len]
                gcodes_in = gcodes_in[buf_len:]
                if type(buf) == str:
                    buf = buf.encode("utf-8")
                    partline = ""
                else:
                    partline = b""
                buf = buf.replace(b"\r", b"")
                if buf:
                    lines = buf.split(b'\n')
                    if gcodes_in:
                        partline = lines.pop()
                    else:
                        while lines and not lines[-1]:
                            lines.pop()
                    gcodes_out.extend(lines)
        self.logger.info('Got %d gcodes to execute.' % len(gcodes_out))
        return gcodes_out
    
    def is_enough_memory(self, size):
        free_mem = self.get_free_memory()
        if free_mem:
            return free_mem > (size * self.MEMORY_STORE_COOF + self.MEMORY_STORE_MARGIN)
        return True # always try if we can't determine a free memory

    def gcodes(self, gcodes):
        start_time = time.monotonic()
        size = 0
        self.logger.debug("Determining gcode format")
        if type(gcodes) == str:
            gcodes = gcodes.encode("utf-8")
        if isinstance(gcodes, io.IOBase):
            try:
                size = os.path.getsize(gcodes.name)
                self.logger.info(f'Loading gcodes file of size: {size/1024/1024}MB' )
            except:
                self.logger.exception('Exception on getting file size:')
            else:
                if not self.is_enough_memory(size):
                    self.parent.register_error(88, "Not enough memory. Cancelling...", is_blocking=True)
                    return False
            gcodes = gcodes.read()
        elif self.is_base64_re.match(gcodes):
            gcodes = base64.b64decode(gcodes)
            return self.unbuffered_gcodes(gcodes)
        self.logger.debug("Start loading gcodes. Determination time:" + str(time.monotonic() - start_time))
        if self.load_gcodes(gcodes) != False: # None is equal to True here
            self.print_start_time = time.monotonic()
            self.logger.info("Done loading gcodes. Time spent:" + str(time.monotonic() - start_time))
            return True
        return False

    @property
    def filename(self):
        return self.displayed_filename

    @filename.setter
    def filename(self, new_filename):
        self.displayed_filename = new_filename

    def set_filename(self, cloud_filename):
        self.full_filename = ""
        self.displayed_filename = ""
        if cloud_filename:
            try:
                self.full_filename = str(cloud_filename)
                self.displayed_filename = filename.get_filename_ascii(os.path.splitext(os.path.basename(self.full_filename))[0])
            except:
                self.logger.warning("Filename is not str: %s" % cloud_filename)

    def get_filename(self):
        return self.displayed_filename 

    def set_estimated_print_time(self, est_print_time):
        estimation = 0
        try:
            estimation = int(est_print_time)
        except (TypeError, ValueError):
            self.logger.warning(f"Given estimated time can not be converted to integer: {est_print_time}")
        self.est_print_time = estimation
        self.print_time_left = estimation
        self.logger.info(f"Setting estimated print duration to {estimation} seconds")
        #self.sum_pause_and_heating_duration = 0
        #self.print_start_time = time.monotonic()
        #self.last_remaining_print_time_get = time.monotonic()

    def set_average_printing_speed(self, speed):
        self.logger.info(f"Average printing speed: {speed}")
        self.average_printing_speed = speed

    def get_remaining_print_time(self, ignore_state=False):
        time_left = 0
        if self.is_printing() or ignore_state:
            if self.est_print_time:
                time_left = int(self.est_print_time - self.est_print_time * self.get_percent() / 100)
            if self.average_printing_speed:
                if not time_left or self.override_clouds_estimations:
                    time_left = int((100 - self.get_percent()) / self.average_printing_speed)
        if self.print_time_left and time_left > self.print_time_left and not self.allow_increase_of_print_time_left:
            time_left = self.print_time_left
        self.logger.info(f"Time left:{time_left}")
        self.print_time_left = time_left
        return time_left

    def get_remaining_print_time_string(self, seconds=None):
        time_string = ""
        if seconds == None:
            seconds = self.get_remaining_print_time()
        if seconds:
            hours, minutes = divmod(seconds // 60, 60)
            if hours:
                time_string += f"{hours} hour"
                if hours != 1:
                    time_string += "s"
            if minutes:
                if time_string:
                    time_string += " "
                time_string += f"{minutes} minute"
                if minutes != 1:
                    time_string += "s"
            if seconds and not hours and not minutes:
                time_string = "less than a minute"
            self.logger.debug("Remaining print time " + time_string)
        return time_string

    #  def get_remaining_print_time(self, ignore_state=False):
    #      if self.is_printing() or ignore_state:
    #          if self.print_time_left and self.est_print_time:
    #              now = time.monotonic()
    #              with self.pause_and_heating_duration_lock:
    #                  if self.heating_start_time:
    #                      self.sum_pause_and_heating_duration += now - self.heating_start_time
    #                      self.heating_start_time = None
    #                  if self.pause_flag:
    #                      if self.last_pause_time:
    #                          self.update_pause_time_and_duration()
    #                  else:
    #                      if self.last_pause_time:
    #                          self.update_unpause_time_and_duration()
    #              #time_progress_relation = int((self.print_start_time + self.sum_pause_and_heating_duration - ) /  print_time_left)
    #              #time_coef = (((self.estimated_print_time - elapsed_time)/self.estimated_print_time) * (1 - self.get_percent() / 100) ** 0.5
    #              time_relation = (self.est_print_time + self.sum_pause_and_heating_duration - self.print_time_left) / self.est_print_time
    #              progress_relation = (100 - self.get_percent()) / 100
    #              if progress_relation:
    #                  speed_coefficient = time_relation / progress_relation
    #              else:
    #                  speed_coefficient = 1
    #              delta_time = now - self.last_remaining_print_time_get
    #              self.last_remaining_print_time_get = now
    #              self.print_time_left -= int(delta_time * speed_coefficient)
    #              if self.print_time_left < 0:
    #                  self.print_time_left = 0
    #              self.logger.info("Remaining print time %s second" % self.print_time_left)
    #              return self.print_time_left 
    #      return 0

    def get_position(self):
        return self.position

    def get_temps(self):
        return self.temps

    def get_target_temps(self):
        return self.target_temps

    def get_percent(self):
        return self.percent

    def get_current_line_number(self):
        return self.current_line_number

    #  def update_pause_time_and_duration(self):
    #      with self.pause_and_heating_duration_lock:
    #          now = time.monotonic()
    #          if self.last_pause_time is not None:
    #              self.sum_pause_and_heating_duration += now - self.last_pause_time
    #          self.last_pause_time = now

    #  def update_unpause_time_and_duration(self):
    #      with self.pause_and_heating_duration_lock:
    #          now = time.monotonic()
    #          self.sum_pause_duration = now - self.last_pause_time 
    #          self.last_pause_time = None

    def pause(self):
        self.pause_flag = True
        #self.update_pause_time_and_duration()

    def unpause(self):
        self.pause_flag = False
        #self.update_unpause_time_and_duration()

    def is_printing(self):
        return self.printing_flag

    def is_paused(self):
        return self.pause_flag

    def is_operational(self):
        return self.operational_flag

    def is_heating(self):
        return self.heating

    def is_bed_not_clear(self):
        return False

    def is_empty(self):
        return False

    def is_maintenance(self):
        return False

    def get_downloading_percent(self):
        return self.parent.downloader.get_percent()

    def get_nonstandart_data(self):
        return {}

    def execute_callback(self, line, success):
        for callback in self.response_callbacks:
            try:
                callback(line, success)
            except:
                self.logger.exception("Exception in callback(%s):" % str(callback))

    def round_temps_list(self, temps_list):
        index = 0
        while index < len(temps_list):
            temps_list[index] = round(temps_list[index], self.TEMPERATURE_SIGNS_AFTER_DOT)
            index += 1
        return temps_list

    def add_response_callback(self, callback_function):
        self.logger.info("Adding callback: %s" % callback_function)
        with self.callbacks_lock:
            if not callback_function in self.response_callbacks:
                self.response_callbacks.append(callback_function)
                self.logger.info("Callback added: %s" % callback_function)

    def del_response_callback(self, callback_function):
        self.logger.info("Removing callback: %s" % callback_function)
        with self.callbacks_lock:
            self.response_callbacks.remove(callback_function)
            self.logger.info("Callback removed: %s" % callback_function)

    def flush_response_callbacks(self):
        with self.callbacks_lock:
            for callback in self.response_callbacks:
                try:
                    self.response_callbacks.remove(callback)
                    self.logger.info("Callback removed: %s" % callback)
                except ValueError:
                    pass

    def init_speed_calculation_thread(self):
        if config.get_settings().get('print_estimation', {}).get('by_print_speed'):
            self.logger.info("Starting print speed calculation thread")
            self.speed_calculation_thread = SpeedCalculationThread(self)
            self.speed_calculation_thread.start()
        else:
            self.logger.info("Print speed calculation is disabled. No thread start")

    def get_jobs(self):
        return {}

    def get_material_names(self):
        return []

    def get_material_volumes(self):
        return []

    def get_material_desc(self):
        """
            primary - default == None
            support
            secondary
            aux - for ink like materials, added to primary(on hp for example)
        """
        return []

    def get_ext(self):
        return {}

    def register_error(self, *args, **kwargs):
        self.parent.register_error(*args, **kwargs)

    def get_estimated_time(self):
        return self.estimated_time

    def get_clouds_job_id(self):
        return self.clouds_job_id

    def get_printers_job_id(self):
        return self.printers_job_id

    def close(self):
        self.stop_flag = True
        if hasattr(self, 'speed_calculation_thread'):
            self.logger.info("Joining estimation thread...")
            self.speed_calculation_thread.join(self.speed_calculation_thread.LOOP_TIME)
            self.logger.info("...estimation thread joined")

    def get_free_memory(self):
        if platforms.PLATFORM in ('rpi', 'linux'):
            try:
                with open('/proc/meminfo') as f:
                    for line in f:
                        if 'MemAvailable' in line:
                            return int(line.split()[1]) * 1024
            except (OSError, IOError, ValueError, IndexError):
                pass


class SpeedCalculationThread(threading.Thread):

    LOOP_STEPS = 100
    LOOP_TIME = 6 # seconds
    SPEEDS_QUEUE_LEN = 24

    def __init__(self, base_sender):
        self.base_sender = base_sender
        self.stop_flag = False
        self.speeds_log = collections.deque(maxlen=self.SPEEDS_QUEUE_LEN)
        self.logger = base_sender.logger.getChild(self.__class__.__name__)
        super().__init__()

    def get_average_speed(self):
        if len(self.speeds_log) == self.SPEEDS_QUEUE_LEN:
            try: #NOTE could use normalize or other formulas instead of average, too increase accuracy
                return sum(self.speeds_log) / self.SPEEDS_QUEUE_LEN
            except IndexError:
                self.logger.exception("Exception while getting average print speed:")

    def run(self):
        printing_counter = 0
        nonprinting_counter = 0
        sleep = self.LOOP_TIME / self.LOOP_STEPS
        last_time = time.monotonic()
        last_percent = 0.0 
        self.logger.info('Entering speed calculation loop')
        while not self.stop_flag and not self.base_sender.stop_flag:
            if self.base_sender.is_operational() and self.base_sender.is_printing():
                printing_counter += 1
            else:
                nonprinting_counter += 1
            if nonprinting_counter >= self.LOOP_STEPS:
                nonprinting_counter = 0
                if self.speeds_log:
                    self.speeds_log.clear()
                    self.base_sender.set_average_printing_speed(0)
            if printing_counter < self.LOOP_STEPS:
                time.sleep(sleep)
            else:
                printing_counter = 0
                if self.base_sender.is_printing() and not self.base_sender.is_paused() and \
                        not self.base_sender.is_heating():
                    percent = self.base_sender.get_percent()
                    delta_time = time.monotonic() - last_time
                    if percent and delta_time:
                        speed = (percent - last_percent) / delta_time
                        self.logger.info(f'Print speed: {speed} %/s')
                        self.speeds_log.append(speed)
                        avg_speed = self.get_average_speed()
                        if avg_speed:
                            self.base_sender.set_average_printing_speed(avg_speed)
                        self.logger.info(f"Delta:{delta_time} Speed:{speed} Avg:{avg_speed}")
                last_percent = self.base_sender.get_percent()
                last_time = time.monotonic()


class FakeParent:

    class FakeApp:

        import user_login

        class FakeLogin(user_login.UserLogin):

            def __init__(self):
                self.logger = logging.getLogger('FakeLogin')
                self.logger.warning("WARNING! Fake parent created! Use only for debug! WARNING!")
                self.login = "3dpos_user"
                self.profiles = self.load_local_printer_profiles()
                #self.init_profiles()

        def __init__(self):
            self.stop_flag = False
            self.user_login = self.FakeLogin()
            self.local_ip = "127.0.0.1"
            self.offline_mode = True

    def __init__(self):
        self.logger = logging.getLogger()
        self.stop_flag = False
        self.error_history = []
        self.printer = None
        self.parent = self.FakeApp()

    def register_error(self, code, message, *args, **kwargs):
        error_record = {'code': code, 'message': message}
        error_record.update(kwargs)
        self.error_history.append(error_record)
        logging.getLogger("FakeParent").warning(f"Reg error:{error_record}")
        if kwargs.get('is_blocking'):
            self.stop_flag = True
            try:
                self.printer.close()
            except:
                pass
