# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import collections
import config
import logging 
import time
import threading
import random

import log

class EmulatedSerialConnection:

    DEFAULT_GCODE_EXEC_SPEED = {"G0" : 0.001, "G1" : 0.001, "G28" : 2}
    IDLE_SLEEP = 0.0001
    HEATING_SPEEDS = [10.0, 30.0]
    HEATING_WAITING_ACK_PERIOD = 1
    HEATING_MISS_THRD = 10 #warning, setting it lower can cause overstep of target temp
    DEFAULT_ERROR_RATE = 0.0
    RESEND_RANGE = 100 
    RECV_PAUSE_ON_NONE = 0.01
    RECV_PAUSE_ON_ACK = 0.00

    def __init__(self, port_name="/dev/null", baudrate=115200, timeout=1, start_dtr=None, logger=None):
        if logger:
            self.logger = logger.getChild('connection')
        else:
            self.logger = logging.getLogger(self.__class__.__name__)
        self.port_name = port_name
        self.baudrate = baudrate
        self.timeout = timeout
        self.verbose = config.get_settings()['verbose']
        self.in_buffer = collections.deque()
        self.out_buffer = collections.deque()
        #self.out_buffer.append('start')
        self.temps = [0.0, 0.0]
        self.target_temps = [0.0, 0.0]
        self.position = [0.0, 0.0, 0.0, 0.0]
        self.wait_heatup = [False, False]
        self.port = port_name
        self.resend_error_rate = self.DEFAULT_ERROR_RATE
        self.error_on_line_number = -1
        self.current_line_number = 0 
        self.exec_thread = threading.Thread(target=self.gcode_execution_loop)
        self.exec_thread.start()

    def set_print_speed_coef(self, coef):
        for key in self.DEFAULT_GCODE_EXEC_SPEED:
            self.DEFAULT_GCODE_EXEC_SPEED[key] = self.DEFAULT_GCODE_EXEC_SPEED[key] * coef

    def parse_arg(self, arg, arg_name, arg_type):
        try:
            parsed = arg_type(arg.replace(arg_name, ''))
        except ValueError:
            error = f'Error: invalid gcode arg:{arg}'
            self.out_buffer.append(error)
            self.logger.warning(error)
            self.logger.debug(f'Arg:{arg} Arg_name:{arg_name}, Arg_type:{arg_type}')
            parsed = None
        return parsed

    def search_and_parse_arg(self, args, arg_name, arg_type):
        for arg in args:
            if arg_name in arg:
                return self.parse_arg(arg, arg_name, arg_type)

    def validate_checksum(self, line, given_checksum): 
        checksum = 0
        for char in line.encode():
            checksum ^= char
        #self.logger.debug(f"Line:{line} Checksum:{given_checksum} Expected:{checksum}")
        if checksum == int(given_checksum):
            return True
        else:
            error = f'Error:checksum mismatch, Last Line: {self.current_line_number}'
            self.out_buffer.append(error)
            self.logger.warning(error)
            return False

    def form_heating_string(self):
        return f"ok T:{self.temps[1]} /{self.target_temps[1]} B:{self.temps[0]} /{self.target_temps[0]}"

    def is_waiting_heatup(self):
        for wait in self.wait_heatup:
            if wait:
                self.out_buffer.append(self.form_heating_string().strip('ok '))
                return True
        return False

    @log.log_exception
    def gcode_execution_loop(self):
        while self.port:
            last_loop_time = time.monotonic()
            #self.logger.debug("HW", self.wait_heatup, 'TT', self.target_temps, 'T', self.temps)
            if self.is_waiting_heatup():
                time.sleep(self.HEATING_WAITING_ACK_PERIOD)
            elif not self.in_buffer:
                time.sleep(self.IDLE_SLEEP)
            else:
                gcode = self.in_buffer.popleft()
                if random.random() < self.resend_error_rate or self.current_line_number == self.error_on_line_number:
                    if self.current_line_number == self.error_on_line_number:
                        self.error_on_line_number = -1
                    resent_from = random.randrange(self.current_line_number-self.RESEND_RANGE, self.current_line_number) 
                    if resent_from < 0:
                        resent_from = 0
                    self.current_line_number = resent_from 
                    self.logger.info('Sending resend from: ' + str(resent_from))
                    self.out_buffer.append('Resend: ' + str(resent_from))
                    self.out_buffer.append('ok')
                    continue
                try:
                    gcode = gcode.decode(errors='ignore')
                except:
                    self.out_buffer.append(f'Error: unable to decode a line')
                gcode_and_check_sum_list = gcode.split('*')
                gcode = gcode_and_check_sum_list[0]
                if len(gcode_and_check_sum_list) > 1:
                    checksum = gcode_and_check_sum_list[1]
                else:
                    checksum = None
                args = gcode.split()
                if not args:
                    continue
                if 'M110' in args:
                    for arg in args:
                        if 'N' in arg:
                            line_number = self.parse_arg(arg, 'N', int)
                            if line_number != None:
                                self.current_line_number = line_number + 1
                                break
                    else:
                        message = f'Invalid line code: {gcode}'
                        self.logger.warning(message)
                        self.out_buffer.append(message)
                    self.out_buffer.append('ok')
                    continue
                elif args[0].startswith('N'):
                    if checksum == None:
                        self.logger.warning(f'Warning:checksum mismatch: {gcode} {checksum}')
                        self.out_buffer.append(f'Error:checksum mismatch, Last Line: {self.current_line_number}')
                        continue
                    elif not self.validate_checksum(gcode, checksum):
                        self.logger.warning(f'Warning:checksum mismatch: {gcode} {checksum}')
                        self.out_buffer.append(f'Error:checksum mismatch, Last Line: {self.current_line_number}')
                        continue
                    line_number = self.parse_arg(args[0], 'N', int)
                    if line_number == None or line_number != self.current_line_number:
                        self.out_buffer.append(f'Line number is not last line number+1. Last line: {self.current_line_number}') #TODO
                        self.out_buffer.append('Resend: ' + str(self.current_line_number))
                        self.out_buffer.append('ok')
                        continue
                    else:
                        self.current_line_number += 1
                    args = args[1:]
                for key in self.DEFAULT_GCODE_EXEC_SPEED:
                    if key and args[0].startswith(key):
                        time.sleep(self.DEFAULT_GCODE_EXEC_SPEED[key])
                        break
                if args[0] == 'M105':
                    self.out_buffer.append(self.form_heating_string())
                elif args[0] == 'M115':
                    self.out_buffer.append(f"ok X:{self.position[0]} Y:{self.position[1]} Z:{self.position[2]} E:{self.position[3]}")
                elif args[0] == 'M104':
                    ttemperature = self.search_and_parse_arg(args, 'S', float)
                    if ttemperature != None:
                        self.target_temps[1] = ttemperature
                        self.logger.info(f"Setting T0 target temp: {ttemperature}")
                        self.out_buffer.append("ok")
                elif args[0] == 'M140':
                    ttemperature = self.search_and_parse_arg(args, 'S', float)
                    if ttemperature != None:
                        self.target_temps[0] = ttemperature
                        self.logger.info(f"Setting bed target temp: {ttemperature}")
                        self.out_buffer.append("ok")
                elif args[0] == 'M109':
                    ttemperature = self.search_and_parse_arg(args, 'S', float)
                    if ttemperature != None:
                        self.target_temps[1] = ttemperature
                        self.wait_heatup[1] = True
                        self.logger.info(f"Setting T0 target temp: {ttemperature}")
                elif args[0] == 'M190':
                    ttemperature = self.search_and_parse_arg(args, 'S', float)
                    if ttemperature != None:
                        self.target_temps[0] = ttemperature
                        self.wait_heatup[0] = True
                        self.logger.info(f"Setting bed target temp: {ttemperature}")
                else:
                    self.out_buffer.append("ok")
            loop_delta = time.monotonic() - last_loop_time  
            for index, ttemp in enumerate(self.target_temps):
                if int(ttemp) > int(self.temps[index]):
                    self.temps[index] = round(min(self.temps[index] + self.HEATING_SPEEDS[index] * loop_delta, ttemp), 0)
                elif int(ttemp) < int(self.temps[index]):
                    self.temps[index] = round(max(self.temps[index] - self.HEATING_SPEEDS[index] * loop_delta, ttemp), 0)
                else:
                    if self.wait_heatup[index]:
                        if index == 0:
                            self.logger.info(f'Heating complete for bed')
                        else:
                            self.logger.info(f'Heating complete for T{index}')
                        self.wait_heatup[index] = False
                        self.out_buffer.append("ok")

    def recv(self, size=None):
        if self.out_buffer:
            if size:
                data = self.out_buffer[0][:1:]
                self.out_buffer[0] = self.out_buffer[0][1:]
            else:
                data = self.out_buffer.popleft()
            if self.verbose and data:
                self.logger.info('RECV: ' + str(data))
            return data.encode()
        elif self.out_buffer == None:
            self.logger.warning('Receive called on closed port')
            time.sleep(1)
        time.sleep(self.RECV_PAUSE_ON_NONE)
        return b""

    def prepare_data(self, data):
        return data.strip() + b"\n"

    def send(self, data, raw=False):
        if not raw:
            data = self.prepare_data(data)
        if self.in_buffer != None:
            if self.verbose:
                self.logger.info('SEND: ' + str(data))
            # if self.in_buffer:
            #     self.logger.warning(f"!!!Got a gcode when in_buffer already got some {len(self.in_buffer)} gcodes\n*10")
            self.in_buffer.append(data)
            return True
        else:
            self.logger.warning('Send called on closed port')
            return False

    def reset(self):
        self.logger.info('Port reset')

    def set_error_rate(self, rate):
        self.resend_error_rate = rate

    def set_error_on_line_number(self, line_number):
        self.error_on_line_number = line_number

    def close(self):
        self.logger.info('Port closed')
        self.in_buffer = []
        self.out_buffer = []
        self.port = ""
