import logging
import threading
import time

try:
    import RPi.GPIO as GPIO
except (ImportError, RuntimeError):
    GPIO = None

import config

class GPIOInterface:

    def __init__(self):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.enabled = False
        self.loop_thread = None
        if config.get_settings().get('gpio').get('enabled'):
            self.init()

    def init(self):
        #TODO refactor events detection
        #it was brutally remade from native GPIO.event_detected(unusable without root)
        self.enabled = False
        if not GPIO:
            self.logger.warning("No RPi.GPIO module - disabling GPIO interface")
            return
        self.logger.info("Enabling GPIO interface")
        trigger_str = config.get_settings().get('gpio').get('trigger')
        if trigger_str == "rising":
            self.trigger = GPIO.RISING
        elif trigger_str == "falling":
            self.trigger = GPIO.FALLING
        elif trigger_str == "both":
            self.trigger = GPIO.BOTH
        else:
            self.logger.warning("Error starting RPi.GPIO: unknown trigger type in config. Should be one of rising, falling, both")
            self.trigger = None
        if not self.trigger:
            self.logger.info("GPIO errir: bad trigger")
        pull_str = config.get_settings().get('gpio').get('pull')
        if pull_str == "up":
            self.pull = GPIO.PUD_UP
        elif pull_str == "down":
            self.pull = GPIO.PUD_DOWN
        else:
            self.pull = None
        self.bouncetime = config.get_settings().get('gpio').get('bouncetime', 200)
        GPIO.setmode(GPIO.BOARD)
        GPIO.setwarnings(False)
        self.active_buttons = []
        self.cancel_callbacks = []
        self.buttons = []
        buttons = config.get_settings()['gpio'].get('buttons', {})
        for button in buttons:
            action_name = button
            pin_number = buttons[button]
            try:
                if self.pull:
                    GPIO.setup(pin_number, GPIO.IN, pull_up_down=self.pull)
                else: #TODO recheck pull_up_down=None here, should it be so when pull_str is "down"?
                    GPIO.setup(pin_number, GPIO.IN)
            except RuntimeError:
                try:
                    GPIO.cleanup()
                except RuntimeError:
                    pass
                self.logger.exception("Exception during GPIO pins setup:")
                return
            else:
                self.buttons.append((pin_number, action_name))
                self.logger.info("GPIO event registered: pin%d=%s" % (pin_number, action_name))
        self.enabled = True
        self.start_loop_thread()
        self.logger.info("GPIO interface init success")

    def start_loop_thread(self):
        if self.loop_thread and self.loop_thread.is_alive():
            self.enabled = False
            self.loop_thread.join()
            self.enabled = True
        self.loop_thread = threading.Thread(target=self.loop)
        self.loop_thread.start()

    def loop(self):
        self.prev_levels = {}
        self.event_detected = {}
        while self.enabled:
            for pin, _ in self.buttons:
                try:
                    level = GPIO.input(pin)
                except RuntimeError:
                    self.logger.warning("GPIO error on reading input of pin: %d" % pin)
                else:
                    if pin in self.prev_levels:
                        prev_level = self.prev_levels[pin]
                        if self.trigger == GPIO.FALLING or self.trigger == GPIO.BOTH:
                            if not level and prev_level:
                                self.event_detected[pin] = True
                        if self.trigger == GPIO.RISING or self.trigger == GPIO.BOTH:
                            if level and not prev_level:
                                self.event_detected[pin] = True
                    else:
                        self.prev_levels[pin] = level
            time.sleep(self.bouncetime/1000)

    def check_events(self, printer_interfaces):
        if self.enabled:
            for pin_num, action_name in self.buttons:
                if self.event_detected.get(pin_num):
                    self.event_detected[pin_num] = False
                    for pi in printer_interfaces:
                        self.run_method(pi, action_name)

    def run_method(self, pi, method_name):
        method = getattr(pi, method_name, None)
        if method:
            try:
                method()
            except Exception:
                self.logger.exception("Exception in GPIO handler")
        else:
            self.logger.error("GPIO event error: no method %s" % method_name)

    def close(self):
        self.enabled = False
        if GPIO:
            GPIO.cleanup()

    def toggle(self):
        if self.enabled:
            self.close()
        else:
            self.init()
        settings = config.get_settings()
        settings['gpio']['enabled'] = self.enabled
        config.Config.instance().save_settings(settings)
