import ipaddress
import json
import http.client
import logging
import socket

import findssh
from findssh.coro import get_hosts as coro_get_hosts
from findssh.threadpool import get_hosts as threadpool_get_hosts
import log


class HTTPDetector:

    PORT = 80
    TIMEOUT = 1
    FINDSSH_SERVICE = "http"
    PATH = "/"
    RESPONSE_FIELD_NAME = "status"

    def __init__(self, parent=None, profile=None):
        self.logger = logging.getLogger('app.' + __name__)
        self.ip = findssh.getLANip()
        self.parent = parent
        self.profile = profile
        self.port = self.PORT
        self.discovered_printers = []
    
    def process_hosts(self, hosts):
        processed_hosts = []
        for host in hosts:
            host_ip = str(host[0])
            if self.parent_stop():
                break
            test_field_contents = self.check_ip(host_ip)
            if test_field_contents:
                processed_hosts.append((host_ip, test_field_contents))
        return processed_hosts

    def check_ip(self, host): 
        self.logger.debug("Detecting host:" + str(host))
        if self.port in (443, 8443):
            connection = http.client.HTTPSConnection(host, self.port, timeout=self.TIMEOUT)
        else:
            connection = http.client.HTTPConnection(host, self.port, timeout=self.TIMEOUT)
        self.logger.debug("Connecting")
        try:
            connection.connect()
            connection.request('GET', self.PATH)
            resp = connection.getresponse()
        except (socket.error, http.client.error) as e:
            self.logger.warning("Error on http detection request: " + str(e))
        else:
            self.logger.debug("Response status code %s" % resp.status)
            if resp.status == 200:
                self.logger.debug("Reading body")
                try:
                    answer_body = resp.read()
                except (socket.error, http.client.error) as e:
                    self.logger.warning("Error on http detection body read: " + str(e))
                else:
                    self.logger.debug("Decoding json")
                    try:
                        decoded = json.loads(answer_body)
                        if type(decoded) != dict:
                            raise TypeError("Response is not json dictionary")
                    except (ValueError, TypeError) as e:
                        self.logger.debug("HTTP detector got non json response: " + str(e))
                    else:
                        self.logger.debug("Checking status dictionary")
                        if isinstance(decoded, dict):
                            test_field_contents = decoded.get(self.RESPONSE_FIELD_NAME)
                        # elif isinstance(decoded, list):
                        #     if decoded:
                        #         test_field_contents = decoded[0]
                        # elif isinstance(decoded, str):
                        #     test_field_contents = decoded
                        else:
                            test_field_contents = None
                        self.logger.debug('Got test field:' + str(test_field_contents))
                        return test_field_contents

    def parent_stop(self):
        return self.parent and self.parent.stop_flag

    def detect(self, already_know_ip=None, non_default_port=None):
        if non_default_port:
            self.port = non_default_port
        else:
            self.port = self.PORT
        if not already_know_ip:
            network = findssh.netfromaddress(self.ip)
            if isinstance(network, ipaddress.IPv6Network):
                hosts = threadpool_get_hosts(network, self.port, self.FINDSSH_SERVICE, self.TIMEOUT)
            else:
                hosts = findssh.runner(coro_get_hosts, network, self.port, self.FINDSSH_SERVICE, self.TIMEOUT)
        else:
            hosts = [(already_know_ip, self.port)]
        self.logger.debug("All responded hosts: " + str(hosts))
        if self.profile:
            VID, PID = self.profile.get('vids_pids', [("XXXX","XXXX")])[0]
        else:
            VID, PID = "XXXX", "XXXX" #dummy values only for debug mode
        discovered_printers = []
        # TODO refactor all the mess with overload of response of proccess_hosts
        for host in self.process_hosts(hosts):
            self.logger.debug("Detected: " + str(host))
            printer = {'VID': VID, 'PID': PID}
            if isinstance(host, tuple):
                host = host[0]
            if isinstance(host, dict):
                printer.update(host)
            elif isinstance(host, str):
                printer['SNR'] = host
                printer['IP'] = host
            if printer not in discovered_printers:
                discovered_printers.append(printer)
        self.discovered_printers = discovered_printers
