# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import time
import json
import select
import socket
import logging
import threading
import os.path

import config
import usb_detect
import paths
import http_rrfw_detect
import http_reperier_detect
import zeroconf_detect
import zeroconf_formlabs_detect
import upnp_detect


# TODO refactor 
# 1) use multithreading or asyncio to scan simultaneously
# 2) decouple starting, running and processing from network detector and other detectors
# 3) refactor to simplify network detector logic and minimize this class

LISTEN_TIMEOUT = 3

class Listening(threading.Thread):

    READ_LENGTH = 1024
    READ_RETRIES = 4

    def __init__(self, port):
        self.logger = logging.getLogger(self.__class__.__name__)
        self.received = {}
        self.addr = None
        self.socket = self.create_socket(port)
        threading.Thread.__init__(self, daemon=True)

    def create_socket(self, port):
        self.logger.debug("Creating listen socket for port %d" % port)
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.bind(('', port))
            sock.settimeout(LISTEN_TIMEOUT)
        except socket.error:
            self.logger.debug("Socket init error. Port:" + str(port), exc_info=True)
        except ValueError:
            self.logger.debug("Not valid port number:" + str(port), exc_info=True)
        else:
            self.logger.debug("...done")
            return sock

    def run(self):
        if self.socket:
            read_retries_left = self.READ_RETRIES
            while read_retries_left:
                read_retries_left -= 1
                try:
                    data, addr = self.socket.recvfrom(self.READ_LENGTH)
                    if addr:
                        if not addr in self.received:
                            self.received[addr] = data
                        # else:
                        #     if self.received[addr] != data: #protection against printer's hiccups 
                        #         self.received[addr] += data
                        #         self.logger.debug(f'Concatenating a broadcast resp from {addr}')
                        #     else:
                        #         self.logger.warning(f'Received what seems to be a duplicate broadcast resp from {addr}')
                except socket.error:
                    break


class NetPrinterScan:

    RETRIES = 1
    RETRY_TIMEOUT = 0.1

    def __init__(self, _, profile):
        self.logger = logging.getLogger('app.' + __name__)
        self.profile = profile
        network_profile = profile['network_detect']
        self.listen_port = network_profile.get('listen_port', None)
        self.broadcast_port = network_profile.get('broadcast_port', None)
        self.target_port = network_profile.get('target_port', None)
        message = network_profile.get('message', '')
        if network_profile.get('json_message'):
            message = json.dumps(message)
        self.message = message
        self.already_detected = []
        self.discovered_printers = []

    def detect(self, already_know_ip=None, non_default_port=None):
        #TODO implement already_know_ip and non_default_port
        if self.broadcast_port and self.listen_port and self.target_port:
            self.listen = Listening(self.listen_port)
            self.listen.start()
            self.broadcast()
            self.listen.join()
        else:        
            error = "Error in config file. Section network_detect of profile: " + str(self.profile)
            self.logger.error(error)

    def create_broadcasting_socket(self):
        self.logger.debug("Creating broadcast socket on port %d" % self.broadcast_port)
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            sock.settimeout(LISTEN_TIMEOUT-1)
            sock.bind(('', self.broadcast_port))
        except socket.error:
            self.logger.debug("Socket init error. Port: %d" % self.broadcast_port, exc_info=True)
        except ValueError:
            self.logger.debug("Not valid port number: %d" % self.broadcast_port, exc_info=True)
        else:
            return sock

    def broadcast(self):
        self.logger.debug("Sending broadcast to port %d" % self.target_port)
        bc_socket = self.create_broadcasting_socket()
        if bc_socket:
            attempt = 1
            while attempt < self.RETRIES + 1:
                try:
                    bc_socket.sendto(self.message.encode('utf-8'), ('255.255.255.255', self.target_port))
                except socket.error:
                    self.logger.debug("Error sending broadcast", exc_info=True)
                    #self.logger.debug("Timeout on port:" + str(self.listen_port))
                attempt += 1
                time.sleep(self.RETRY_TIMEOUT)
                self.listen.join()
                for printer_addr in self.listen.received:
                    self.process_response(self.listen.received[printer_addr], printer_addr)
                    break
            self.logger.debug("Done listening to port " + str(self.listen_port))
            bc_socket.close()

    def process_response(self, response, addr):
        if addr not in self.already_detected:
            self.already_detected.append(addr)
            printer = {'IP': addr[0], 'port': addr[1]}
            if self.profile['network_detect']['json_response']:
                try:
                    response = json.loads(response)
                except (TypeError, ValueError):
                    self.logger.debug('Response from printer should be valid json. Its malformed or not json')
                else:
                    ip_field = self.profile['network_detect']['IP_field']
                    iserial_field = self.profile['network_detect']['SNR_field']
                    vid_field = self.profile['network_detect']['VID_field']
                    pid_field = self.profile['network_detect']['PID_field']
                    printer['IP'] = response[ip_field]
                    if addr[0] != printer['IP']:
                        self.logger.warning("Detected printer IP didn't match with IP field in response")
                    printer['SNR'] = response[iserial_field]
                    printer['VID'] = usb_detect.BaseDetector.format_vid_or_pid(response[vid_field])
                    printer['PID'] = usb_detect.BaseDetector.format_vid_or_pid(response[pid_field])
            if printer not in self.discovered_printers:
                self.discovered_printers.append(printer)

    def close(self):
        pass # no need to close anything here, but method is still should exist until Awaitable implementation


class NetworkDetector:

    STORAGE_FILE_PATH = os.path.join(paths.CURRENT_SETTINGS_FOLDER, 'detected_printers.json')

    def __init__(self, parent=None):
        self.parent = parent
        self.logger = logging.getLogger(self.__class__.__name__)
        self.detected_printers = []
        self.printers_add_by_ip = []
        self.load_printers_list()
        self.network_detector_run_flag = False

    def refresh_on_next_loop(self):
        self.network_detector_run_flag = True

    def parent_stop(self):
        return self.parent and self.parent.stop_flag

    def get_all_network_profiles(self):
        return [x for x in config.get_profiles() if x.get('network_detect')]

    def get_unique_network_profiles(self):
        profiles = []
        network_detects = []
        for profile in self.get_all_network_profiles():
            if profile['network_detect'] not in network_detects:
                profiles.append(profile)
                network_detects.append(profile['network_detect'])
        return profiles

    def get_printers_list(self):
        if self.network_detector_run_flag:
            self.logger.info('Scanning for network printers...')
            self.detected_printers = []
            for profile in self.get_unique_network_profiles():
                if self.parent_stop():
                    return []
                scanner_name = profile.get('network_detect', {}).get('scanner')
                scanner_class = SCANNERS.get(scanner_name)
                if not scanner_class:
                    self.logger.error(f'Unknown scanner name: {scanner_name}')
                    continue
                scanner = scanner_class(self.parent, profile)
                scanner.detect()
                self.logger.debug('Scanner:\t%s\t\%s' % (scanner, scanner.discovered_printers))
                self.detected_printers.extend(scanner.discovered_printers)
            self.logger.info('Discovered printers:\n%s' % self.detected_printers)
            self.save_printers_list()
            self.network_detector_run_flag = False
        return self.detected_printers + self.printers_add_by_ip

    def printers_list_with_profiles(self):
        printers_list = []
        for info in self.get_printers_list():
            for pi in getattr(self.parent, "printer_interfaces", []):
                if pi.usb_info == info:
                    printers_list.append({"usb_info": info, "profile": pi.printer_profile})
                    break
            else:
                printers_list.append({"usb_info": info, "profile": {}})
        return printers_list

    def add_network_printer(self, printer_ip, printer_type, printer_port=None, run_detector=True):
        for profile in self.get_all_network_profiles():
            if printer_type == profile.get("name"):
                scanner_name = profile.get('network_detect', {}).get('scanner')
                if run_detector and scanner_name: #disable scan for birdwing printers
                    if scanner_name in SCANNERS:
                        scanner_class = SCANNERS[scanner_name]
                        scanner = scanner_class(self.parent, profile)
                        scanner.detect(already_know_ip=printer_ip, non_default_port=printer_port)
                        for printer in scanner.discovered_printers:
                            if printer not in self.printers_add_by_ip and not printer in self.detected_printers:
                                self.printers_add_by_ip.append(printer)
                        self.save_printers_list()
                        break
                    else:
                        self.logger.warning(f'Unknown scanner: {scanner_name}')
                else:
                    printer = {
                        'IP': printer_ip,
                        'SNR': printer_ip,
                        'VID': profile["vids_pids"][0][0],
                        'PID': profile["vids_pids"][0][1]
                    }
                    if printer not in self.printers_add_by_ip:
                        self.printers_add_by_ip.append(printer)
                        self.save_printers_list()
                    break

    def forget(self, usb_info):
        for group in (self.detected_printers, self.printers_add_by_ip):
            for printer in group:
                if printer == usb_info:
                    group.remove(printer)
        self.save_printers_list()

    def save_printers_list(self):
        file_path = self.STORAGE_FILE_PATH
        try:
            with open(file_path, 'w') as f:
                data = {'detected': self.detected_printers, 'added_by_ip': self.printers_add_by_ip}
                json_config = json.dumps(data, indent = 4, separators = (',', ': '))
                f.write(json_config)
        except Exception as e:
            self.logger.error("Error writing data to %s: %s" % (file_path, str(e)))

    def load_printers_list(self):
        file_path = self.STORAGE_FILE_PATH
        try:
            with open(file_path, 'r') as f:
                data = f.read()
                self.logger.info("Stored printers list: %s" % data)
                printers_list = json.loads(data)
                self.detected_printers = printers_list.get('detected', [])
                self.printers_add_by_ip = printers_list.get('added_by_ip', [])
        except FileNotFoundError:
            pass
        except Exception as e:
            self.logger.error("Error reading data from %s: %s" % (file_path, str(e)))


SCANNERS = {'broadcast': NetPrinterScan,\
            'rrfw' : http_rrfw_detect.HTTPRRFWDetector,\
            'repetier' : http_reperier_detect.HTTPRepetierDetector,\
            'zeroconf' : zeroconf_detect.ZeroconfDetector,\
            'zeroconf_formlabs' : zeroconf_formlabs_detect.ZeroconfDetector,\
            'upnp' : upnp_detect.UPNPDetector}


if __name__ == '__main__':
    from pprint import pprint

    logging.basicConfig(level=logging.DEBUG)
    nd = NetworkDetector()
    nd.refresh_on_next_loop()
    print("Detected network printers:")
    nd.get_printers_list()
