import logging
import serial_connection
import socket
import telnetlib
import threading
import time

import config


class TelnetConnection(serial_connection.SerialConnection):

    DEFAULT_TIMEOUT = 1
    FAIL_WAIT = 1
    RETRIES = 3
    ENDOFLINE = b'\n'

    def __init__(self, ip, port, timeout=None):
        self.logger = logging.getLogger(__name__)
        self.ip = ip
        self.port = port
        if timeout is None:
            self.timeout = self.DEFAULT_TIMEOUT
        else:
            self.timeout = timeout
        self.recv_lock = threading.RLock()
        self.send_lock = threading.RLock()
        self.verbose = config.get_settings()['verbose']
        self.connection = None
        self.connect()

    def connect(self):
        try:
            self.connection = telnetlib.Telnet(self.ip, self.port, timeout=self.timeout)
        except OSError as e:
            self.logger.error("Error connecting to %s:%s - %s" % (self.ip, self.port, e))

    def recv(self, timeout=None):
        if not timeout:
            timeout = self.timeout
        fails = 0
        while fails < self.RETRIES:
            if not self.connection:
                self.logger.warning("Can't send - no connection to host")
                self.connect()
            data = None
            try:
                with self.recv_lock:
                    if self.connection:
                        data = self.connection.read_until(self.ENDOFLINE, timeout=timeout)
                    else:
                        raise OSError('No connection')
                if self.verbose:
                    self.logger.info("RECV: %s" % data)
            except socket.timeout:
                self.logger.warning("Timeout while reading telnet from %s:%s" % (self.ip, self.port))
                return data
            except (OSError, EOFError) as e:
                self.logger.warning("Can't read telnet from %s:%s. Error:%s" % (self.ip, self.port, e))
            except Exception as e:
                self.logger.warning("Unexpected error while reading from telnet %s:%s" % (self.ip, self.port))
            else:
                return data
            fails += 1
            time.sleep(self.FAIL_WAIT)
            self.close()

    def send(self, data, raw=False):
        if not raw:
            data = self.prepare_data(data)
        try:
            with self.send_lock:
                if self.connection:
                    self.connection.write(data)
                    if self.verbose:
                        self.logger.info("SEND: %s" % data.strip())
        except OSError as e:
            self.logger.warning("Can't write to telnet on %s:%s. Error:%s" % (self.ip, self.port, str(e)))
        except Exception as e:
            self.logger.warning("Unexpected error while writing to telnet %s:%s" % (self.ip, self.port))
        else:
            return True

    def reset(self):
        self.connection.close()

    def close(self):
        self.logger.info("Closing telnet connection %s" % self.ip)
        with self.recv_lock:
            if self.connection:
                self.connection.close()
                self.connection = None
