# Copyright 3D Control Systems, Inc. All Rights Reserved 2017-2019.
# Built in San Francisco.

# This software is distributed under a commercial license for personal,
# educational, corporate or any other use.
# The software as a whole or any parts of it is prohibited for distribution or
# use without obtaining a license from 3D Control Systems, Inc.

# All software licenses are subject to the 3DPrinterOS terms of use
# (available at https://www.3dprinteros.com/terms-and-conditions/),
# and privacy policy (available at https://www.3dprinteros.com/privacy-policy/)

import collections
import logging
import re
import time
import threading
import string

import config
import log
from base_sender import BaseSender
from serial_connection import SerialConnection
from emulated_serial_connection import EmulatedSerialConnection


class Sender(BaseSender):

    CONNECTION_CLASS = SerialConnection
    BUFFER_CLASS = collections.deque

    OK_TIMEOUT = 180 
    ERROR_REPORT_TIMEOUT = 3600
    CONNECTION_TIMEOUT = 4
    TEMP_REQUEST_PERIOD = 3
    PRINTER_INIT_WAIT = 3
    PRINT_JOIN_TIMEOUT = 12
    STOP_JOIN_TIMEOUT = 6
    PAUSE_ON_ERROR_READING_PORT = 0.0001 # should be much less than gcode rate 
    RECONNECTION_RESET_SLEEP = 2
    CONNECTION_RETRIES_PER_BAUDRATE = 2
    CONNECTION_LINE_READ_ATTEMPTS = 6
    MAX_VALID_RESEND_RANGE = 4096
    VALID_START_TIMEOUT = 30
    RATING_TIME = 60 # should be 1 min exactly, due to "/min" in log messages

    GET_TEMP_GCODE = b'M105'
    GET_POSITION_GCODE = b'M114'
    GET_FIRMWARE_VERSION_GCODE = b'M115'
    BOARD_RESET_GCODE = b'M999'

    OK = b'ok'
    ENQUEUEIN_BYTES = b'enqueuein'
    START = b'start'
    POSITIVE_ACKS = [START, b'Grbl', b'grbl', b'echo'] + [OK]
    RS = b'rs'
    CRASH_RECOVER_MESSAGE = b'CRASH_RECOVER'
    BUSY_PROCESSING_MESSAGE = b'busy: processing'
    BUSY_PAUSED_MESSAGE = b'busy: paused for user'
    STORED_SETTINGS_RETRIEVED = b"Stored settings retrieved" 
    FIRMWARE_MESSAGE = b"Firmware"
    FIRMWARE_MESSAGE_LOWERCASE = b"firmware"
    DEBUG_MESSAGE = b'DEBUG'
    ERROR_MESSAGE = b'Error'
    ERROR_MESSAGE_LOWERCASE = b'error'
    WARNING_MESSAGE = b'Warning'
    WARNING_MESSAGE_LOWERCASE = b'warning'
    DOUBLE_EXCL_MARK = b"!!"
    RESEND_REQUEST_PREFIXES = [b'resend', RS]
    TEMPERATURE_ACKS_PREFIXES = [b'T:', b'B:', b'T0:', b'T1:', b'T2:']

    DEFAULT_PAUSE_AT_GCODES_LIST = [b"M0", b"M1", b"M25", b"M226", b"M601", b"@pause"]
    DEFAULT_RESUME_AT_GCODES_LIST = [b"M24", b"M602"]
    DRY_RUN_COMMAND = b"M111 S8"
    IGNORE_PAUSE_MAGIC = b'3DPrinterOS pass pause'

    PRINTABLE_BYTES = string.printable.encode()

    PAUSE_LIFT_HEIGHT = 5 # in mm
    PAUSE_EXTRUDE_LENGTH = 7 # in mm

    DEFAULT_ON_PAUSE_GCODES = (b"G1 Z+%.3f E-%.3f" % (PAUSE_LIFT_HEIGHT, PAUSE_EXTRUDE_LENGTH), )
    DEFAULT_ON_RESUME_GCODES = (b"G1 Z-%.3f E+%.3f" % (PAUSE_LIFT_HEIGHT, PAUSE_EXTRUDE_LENGTH), )

    def __init__(self, parent, usb_info, profile):
        super().__init__(parent, usb_info, profile)
        #TODO move CONNECTION_CLASS from class attributes to object attributes, but first refactor telnet sender
        if usb_info.get('EMU'):
            self.CONNECTION_CLASS = EmulatedSerialConnection
        self.define_regexps()
        self.operational_flag = False
        self.correct_baudrate = None
        self.heating_emulator = None
        self.connection = None
        self.was_in_relative_before_pause = False
        self.firmware_info = ""
        self.read_thread = None
        self.print_thread = None
        self.temperature_request_thread = None
        self.ready_for_command = threading.Semaphore()
        self.thread_start_lock = threading.RLock()
        self.pause_gcodes_list = profile.get("pause_at_gcodes", self.DEFAULT_PAUSE_AT_GCODES_LIST)
        self.resume_gcodes_list = profile.get("resume_at_gcodes", self.DEFAULT_RESUME_AT_GCODES_LIST)
        self.creation_time = time.monotonic()
        self.on_connection_gcodes = self.preprocess_gcodes(profile.get("on_connection_gcodes", profile.get("end_gcodes", [])))
        self.on_pause_gcodes = self.preprocess_gcodes(profile.get("on_pause_gcodes", self.DEFAULT_ON_PAUSE_GCODES))
        self.on_resume_gcodes = self.preprocess_gcodes(profile.get("on_resume_gcodes", self.DEFAULT_ON_RESUME_GCODES))
        # if profile.get("cooldown_on_pause"):
        #     self.on_pause_gcodes.extend([b"M190 S0", b"M109 S0", b"M109 T1 S0"])
        if profile.get("heatup_on_resume"):
            self.on_resume_gcodes.extend([b"M190 S{bed_temp}", b"M109 S{t0_temp}", b"M109 T1 {t1_temp}"])
        self.disable_resend_max = config.get_settings().get('disable_resend_max', False)
        self.init() #TODO split init and constructor and rename init to something more suitable

    def sync_ready_for_command(self, release=True):
        while not self.stop_flag and self.ready_for_command.acquire(blocking=False):
            self.logger.info('Syncing ready for command semaphore...')
        if release:
            self.ready_for_command.release()

    def init(self):
        if self.connect():
            self.last_line_sent = b''
            self.in_relative_pos_mode = False
            self.heating = False
            self.print_thread = None
            self.thread_start_lock = threading.Lock()
            self.ready_for_command = threading.Semaphore()
            self.paused_for_user = False
            self.stop_temp_requesting_flag = False
            self.read_thread = threading.Thread(target=self.reading, name="Read thread", daemon=True)
            self.read_thread.start()
            time.sleep(self.PRINTER_INIT_WAIT)
            self.sync_ready_for_command(True)
            if self.heating_emulator:
                self.heating_emulator.disable(False)
            if not self.profile.get("no_homing_on_connect"):
                self.send_homing_gcodes()
            self.start_temp_requesting()
        else:
            raise RuntimeError("Error: printer is not responding")

    def connect(self):
        if not self.usb_info.get('COM'):
            raise RuntimeError("No serial port detected for serial printer")
        self.logger.info('Baudrates list for %s : %s' % (self.profile['name'], self.profile['baudrate']))
        for baudrate in self.profile['baudrate']:
            if type(baudrate) != int:
                try:
                    baudrate = int(baudrate)
                except (TypeError, ValueError):
                    self.logger.error(f'Error: not integer baudrate {baudrate} in profile ' + self.profile['alias'])
                    continue
            if self.correct_baudrate: #FIXME too ugly
                baudrate = self.correct_baudrate
            if self.connection:
                self.connection.close()
            self.logger.info("Connecting at baudrate %d" % baudrate)
            self.connection = self.CONNECTION_CLASS(self.usb_info['COM'], baudrate, logger=self.logger)
            if self.connection.port:
                retries = self.CONNECTION_RETRIES_PER_BAUDRATE
                while retries and (not self.parent or not self.parent.stop_flag) and not self.stop_flag:
                    connection_result = self.try_to_communicate()
                    if connection_result:
                        self.correct_baudrate = baudrate
                        self.logger.info("Successful connection to %s: %s" % (self.profile['alias'], str(self.usb_info)))
                        return True
                    retries -= 1
            if not self.profile.get("no_DTR"):
                self.connection.reset()
                time.sleep(self.RECONNECTION_RESET_SLEEP)
            self.logger.warning("Error connecting to printer at %d" % baudrate)
            self.connection.close()

    def try_to_communicate(self):
        if self.connection.send(b''): # end a possible leftover from last connection, etc
            if self.connection.send(self.GET_TEMP_GCODE):
                line_read_attempts = self.CONNECTION_LINE_READ_ATTEMPTS
                recv_text = b""
                while not self.stop_flag and line_read_attempts:
                    recv_line = self.connection.recv()
                    if recv_line:
                        recv_text += recv_line
                        self.logger.info('On connection recv: ' + recv_line.decode(errors="ignore"))
                    if self.BOARD_RESET_GCODE in recv_text:
                        self.logger.info("Printer's board requested reset. Resetting...")
                        return False
                    for ack in self.POSITIVE_ACKS:
                        if ack in recv_text:
                            self.logger.info("Printer is online")
                            return True
                    for ack in self.TEMPERATURE_ACKS_PREFIXES:
                        if ack in recv_text:
                            self.logger.info("Printer is online, but in a heating mode.")
                            self.heating = True
                            return True
                    line_read_attempts -= 1
        time.sleep(0.5)

    def analyze_sent_line(self, line):
        self.last_line_sent = line
        if b'M109' in line or b'G10\n' in line or b'G10 ' in line:
            tool_match = re.search(b'T(\d+)', line)
            tool = 1
            if tool_match:
                try:
                    tool = int(tool_match.group(1)) + 1
                except ValueError:
                    self.logger.error("Invalid temp formant in line:%s" % line)
            temp_match = re.search(b'S([\d\.]+)', line)
            if temp_match:
                if not tool >= len(self.target_temps):
                    try:
                        self.target_temps[tool] = float(temp_match.group(1))
                    except ValueError:
                        self.logger.error("Invalid temp formant in line:%s" % line)
        elif b'M190' in line:
            temp_match = re.search(b'S([\d\.]+)', line)
            if temp_match:
                try:
                    self.target_temps[0] = float(temp_match.group(1))
                except ValueError:
                    self.logger.error("Invalid temp formant in line:%s" % line)
        elif b'G90' in line:
            self.in_relative_pos_mode = False
        elif b'G91' in line:
            self.in_relative_pos_mode = True
        if not self.GET_TEMP_GCODE in line and not self.GET_POSITION_GCODE in line:
            self.heating = b'M109' in line or b'M190' in line or b'M116' in line or b'M98' in line or b"G10 " in line

    def send_now(self, line_or_lines_list):
        with self.thread_start_lock:
            if not self.print_thread or not self.print_thread.is_alive():
                self.print_thread = PrintThread(self, name="SendNowThread")
                self.add_lines_to_send_now_buffer(line_or_lines_list)
                self.print_thread.start()
            else:
                self.add_lines_to_send_now_buffer(line_or_lines_list)

    def add_lines_to_send_now_buffer(self, line_or_lines_list):
        if isinstance(line_or_lines_list, str):
            line_or_lines_list = line_or_lines_list.encode('utf-8')
        if isinstance(line_or_lines_list, bytes):
            self.print_thread.send_now_buffer.append(line_or_lines_list)
        else:
            self.print_thread.send_now_buffer.extend(line_or_lines_list)

    def send_homing_gcodes(self):
        with self.thread_start_lock:
            self.logger.info("Starting homing thread")
            gcodes = []
            gcodes.extend(self.on_connection_gcodes)
            gcodes.append(self.GET_FIRMWARE_VERSION_GCODE)
            self.print_thread = PrintThread(self, name="HomingThread")
            self.print_thread.send_now_buffer.extend(gcodes)
            self.print_thread.start()
            self.print_thread.join(self.PRINT_JOIN_TIMEOUT)
            if self.print_thread.is_alive():
                self.logger.warning("Timeout waiting for homing thread to end")

    def define_regexps(self):
        digits_re = b"(-?[\d\.]+)"
        self.temp_re = re.compile(b'T:dr /dr B:dr /dr'.replace(b'dr', digits_re))
        self.position_re = re.compile(b'X:dr.?Y:dr.?Z:dr.?E:dr.*'.replace(b'dr', digits_re))
        self.wait_tool_temp_re = re.compile(b'T\d?:dr'.replace(b'dr', digits_re))
        self.wait_platform_temp_re = re.compile(b'B:dr'.replace(b'dr', digits_re))
        self.wait_tool_target_temp_re = re.compile(b'T\d?:. /dr'.replace(b'dr', digits_re))
        self.wait_platform_target_temp_re = re.compile(b'B\d?:. /dr'.replace(b'dr', digits_re))
        self.bed_re = re.compile(b'B:dr /dr'.replace(b'dr', digits_re))
        self.current_tool_re = re.compile(b'T:dr /dr'.replace(b'dr', digits_re))
        self.first_tool_re = re.compile(b'T0:dr /dr'.replace(b'dr', digits_re))
        self.second_tool_re = re.compile(b'T1:dr /dr'.replace(b'dr', digits_re))
        self.targetless_temp_re = re.compile(b'T:dr\sB:dr'.replace(b'dr', digits_re))
        self.resend_number_re = re.compile(b'(\d+)')
        self.ignore_pause_resume_re = re.compile(b'M[0|1]\s+[PS]\d+')

    @log.log_exception
    def printer_answer_rate_counting(self):
        while not self.stop_flag:
            with self.rate_lock:
                self.oks_rate = 0
                self.empty_rate = 0
                self.busy_rate = 0
                self.wait_rate = 0
                self.none_rate = 0 
                self.temp_rate = 0 
                self.resend_rate = 0
                self.error_rate = 0 
                self.strange_rate = 0 
            counter = self.RATING_TIME
            while counter > 0: # should have > 0, because float is abomination
                if self.stop_flag:
                    return
                counter -= 0.1
                time.sleep(0.1)
            if self.is_printing():
                self.logger.info(f"Ok/min:{self.oks_rate}\n"
                                 f"Empty/min:{self.empty_rate}\n"
                                 f"Busy/min:{self.busy_rate}\n"
                                 f"Wait/min:{self.wait_rate}\n"
                                 f"None/min:{self.none_rate}\n"
                                 f"Temp/min:{self.temp_rate}\n" 
                                 f"Resend/min:{self.resend_rate}\n"
                                 f"Warning/min:{self.error_rate}\n"
                                 f"Strange/min:{self.strange_rate}")

    @log.log_exception
    def reading(self):
        self.rate_lock = threading.Lock()
        self.none_rate = 0 
        self.empty_rate = 0
        self.wait_rate = 0
        self.oks_rate = 0
        self.temp_rate = 0 
        self.resend_rate = 0
        self.error_rate = 0 
        self.busy_rate = 0
        self.strange_rate = 0
        self.asnwer_rate_thread = threading.Thread(target=self.printer_answer_rate_counting, daemon=True)
        self.asnwer_rate_thread.start()
        self.firmware_crash_recover_happened = False
        last_reply_time = time.monotonic()
        # printer response parsing start
        while not self.stop_flag:
            line = self.connection.recv()
            if line is None: # None means error reading from printer, but "" is ok
                with self.rate_lock:
                    self.none_rate += 1
                self.operational_flag = False
                time.sleep(self.PAUSE_ON_ERROR_READING_PORT)
                if self.paused_for_user:
                    self.parent.register_error(299, "Unexpected printer's firmware reboot", is_blocking=True)
            else:
                line = line.strip()
                if not line:
                    #Empty line response means the printer just got nothing to say at the moment ), but is operational
                    with self.rate_lock:
                        self.empty_rate += 1
                else: 
                    callback_good = False
                    last_reply_time = time.monotonic()
                    if line == b"wait":
                        with self.rate_lock:
                            self.wait_rate += 1
                        self.operational_flag = True
                    elif line == self.START and self.temperature_request_thread and not self.firmware_crash_recover_happened:
                        if self.firmware_crash_recover_happened:
                            self.firmware_crash_recover_happened = False
                        else:
                            self.parent.register_error(299, "Unexpected printer's firmware reboot", is_blocking=True)
                    elif line.startswith(self.OK) or line.startswith(self.RS): 
                        self.operational_flag = True
                        if self.paused_for_user:
                            self.paused_for_user = False
                        callback_good = not self.check_temperature(line)
                        self.ready_for_command.release()
                        with self.rate_lock:
                            self.oks_rate += 1
                    elif self.CRASH_RECOVER_MESSAGE in line:
                        self.firmware_crash_recover_happened = True
                        self.ready_for_command.release()
                    elif self.BUSY_PROCESSING_MESSAGE in line:
                        with self.rate_lock:
                            self.busy_rate += 1
                        self.logger.info('Printer busy: %s' % line)
                    elif self.check_position(line):
                        self.logger.info(f"Position: {self.position}")
                    elif self.parse_waiting_temperature_updates(line):
                        with self.rate_lock:
                            self.temp_rate += 1
                        self.operational_flag = True
                    elif self.ENQUEUEIN_BYTES in line:
                        self.logger.error("Printer executing: %s" % line.decode("utf-8", errors='ignore'))
                    elif self.check_for_resend_requests(line):
                        with self.rate_lock:
                            self.resend_rate += 1
                        self.logger.info('Last sent line: %s' % self.last_line_sent)
                    elif self.DOUBLE_EXCL_MARK in line or self.ERROR_MESSAGE in line or self.WARNING_MESSAGE in line or self.ERROR_MESSAGE_LOWERCASE in line or self.WARNING_MESSAGE_LOWERCASE in line:
                        with self.rate_lock:
                            self.error_rate += 1
                        self.logger.info(f'Warning/Error: {line}. As response to line: {self.last_line_sent}')
                        for warning in (self.BOARD_RESET_GCODE, b'disabling all for safety!', b"kill", b"halt" in line, b"critical" in line):
                            if warning in line.lower():
                                is_blocking = True
                                self.operational_flag = False
                                break
                        else:
                            is_blocking = False
                        self.parent.register_error(201, "Printer: %s" % line, is_blocking=is_blocking)
                        # if line.endswith(self.OK): # a hacky fix for Duet3D bug of line ending with ok and lost \n
                        #     self.ready_for_command.release()
                    elif self.FIRMWARE_MESSAGE in line or self.FIRMWARE_MESSAGE_LOWERCASE in line:
                        self.firmware_info = line.decode('utf-8', errors='ignore')
                        self.logger.info("Firmware info: " + self.firmware_info)
                        if self.paused_for_user:
                            self.paused_for_user = False
                    elif line.startswith(self.DEBUG_MESSAGE):
                        self.logger.debug("Debug %s" % line)
                    elif self.BUSY_PAUSED_MESSAGE in line:
                        self.logger.warning('Received: paused for user')
                        self.operational_flag = True
                        if not self.paused_for_user:
                            self.paused_for_user = True
                            self.parent.register_error(252, 'Received: Paused for user', is_info=True)
                    elif self.STORED_SETTINGS_RETRIEVED in line:
                        if self.paused_for_user:
                            self.paused_for_user = False
                    else:
                        with self.rate_lock:
                            self.strange_rate += 1
                        self.log_strange_acks(line)
                    if self.response_callbacks:
                        self.execute_callback(line, callback_good)
            if last_reply_time + self.OK_TIMEOUT < time.monotonic():
                last_reply_time = time.monotonic()
                if not self.paused_for_user and not self.heating and not self.is_paused():
                    message = "Warning! Timeout while waiting for ok. Assuming printer's firmware malfunction."
                    self.parent.register_error(251, message, is_blocking=False)
                    last_error_report_time = time.monotonic()
                    self.logger.warning(f"Attempting to recover printing process by stop to wait for ok.\n\
                            Last line received: {line}. Last line send: {self.last_line_sent}")
                    self.stop_temp_requesting()
                    if self.temperature_request_thread and self.temperature_request_thread.is_alive():
                        self.temperature_request_thread.join()
                    self.flush_send_now_buffer()
                    self.sync_ready_for_command(True)
                    self.start_temp_requesting()
        self.asnwer_rate_thread.join(1)
        self.logger.debug("Reading thread exit")

    def log_strange_acks(self, line):
        self.logger.info('Received: %s' % line)

    def start_temp_requesting(self):
        self.stop_temp_requesting_flag = False
        if self.temperature_request_thread and self.temperature_request_thread.is_alive():
            self.logger.warning("Can't start another temperature requesting thread is alive")
        else:
            self.temperature_request_thread = threading.Thread(target=self.temperature_requesting, name="TemperatureThread")
            self.temperature_request_thread.start()

    def stop_temp_requesting(self):
        self.stop_temp_requesting_flag = True
        try:
            if self.temperature_request_thread and self.temperature_request_thread.is_alive():
                self.temperature_request_thread.join(self.TEMP_REQUEST_PERIOD)
                if self.temperature_request_thread.is_alive():
                    self.logger.warning("Timeout waiting for temperature requesting thread to end")
        except (AttributeError, ValueError, RuntimeError):
            pass

    # TODO remove this thread, instead in idle spawn a print thread named idle thread and make threads add M105 in their buffer themselves. 
    @log.log_exception
    def temperature_requesting(self):
        self.logger.info("Starting temperature requesting loop")
        number_of_steps = 100
        sleep_step = self.TEMP_REQUEST_PERIOD / float(number_of_steps)
        while not self.stop_flag and not self.stop_temp_requesting_flag:
            # don't send M105 if M105 is already in buffer or in blocking heating 
            if not getattr(self.print_thread, 'send_now_buffer', False) and not self.heating:
                self.send_now(self.GET_TEMP_GCODE)
            sleep_step_count = number_of_steps
            while not self.stop_flag and sleep_step_count and not self.stop_temp_requesting_flag:
                time.sleep(sleep_step)
                sleep_step_count -= 1
        self.logger.info("Temperature requesting loop finish")

    def parse_temperature(self, line):
        self.logger.debug(line)
        match = self.temp_re.search(line)
        if match:
            try:
                tool_temp = float(match.group(1))
                tool_target_temp = float(match.group(2))
                platform_temp = float(match.group(3))
                platform_target_temp = float(match.group(4))
                self.temps = [platform_temp, tool_temp]
                self.target_temps = [platform_target_temp, tool_target_temp]
                if self.heating_emulator:
                    self.heating_emulator.check_temperatures()
                return True
            except Exception:
                self.logger.exception("Error parsing temperatures from line:" + str(line))
        else:
            got_temps = False
            temps = []
            target_temps = []
            bed_match = self.bed_re.search(line)
            first_tool_match = self.first_tool_re.search(line)
            second_tool_match = self.second_tool_re.search(line)
            try:
                if bed_match:
                    temps.append(float(bed_match.group(1)))
                    target_temps.append(float(bed_match.group(2)))
                    got_temps = True
                else:
                    temps.append(0.0)
                    target_temps.append(0.0)
                if first_tool_match:
                    temps.append(float(first_tool_match.group(1)))
                    target_temps.append(float(first_tool_match.group(2)))
                    got_temps = True
                else:
                    temps.append(0.0)
                    target_temps.append(0.0)
                if second_tool_match:
                    temps.append(float(second_tool_match.group(1)))
                    target_temps.append(float(second_tool_match.group(2)))
                    got_temps = True
                if got_temps:
                    self.temps = temps
                    self.target_temps = target_temps
                    return True
                else:
                    targetless_match = self.targetless_temp_re.search(line)
                    if targetless_match:
                        self.temps = [(float(targetless_match.group(2))), (float(targetless_match.group(1)))]
                        return True
            except Exception:
                self.logger.exception("Error parsing temperatures from line:" + str(line))
        return False

    def parse_waiting_temperature_updates(self, line):
        is_temperatures_line = False
        if line.startswith(b'T') or line.startswith(b'B:'): # performance optimization 
            match = self.wait_platform_temp_re.search(line)
            if match:
                try:
                    self.temps[0] = float(match.group(1))
                    is_temperatures_line = True
                except ValueError:
                    self.logger.error("Invalid temp formant in line:%s" % line)
            match = self.wait_tool_temp_re.search(line)
            if match:
                try:
                    self.temps[1] = float(match.group(1))
                    is_temperatures_line = True
                except ValueError:
                    self.logger.error("Invalid temp formant in line:%s" % line)
        return is_temperatures_line

    def check_for_resend_requests(self, line):
        for prefix in self.RESEND_REQUEST_PREFIXES:
            if line.lower().startswith(prefix):
                match = self.resend_number_re.search(line)
                if match:
                    failed_line_number = int(match.group(1))
                    self.logger.info("Request to resend line N%d" % failed_line_number)
                    try:
                        with self.print_thread.resend_lock:
                            if not self.disable_resend_max and self.print_thread.lines_sent - self.print_thread.n_offset > failed_line_number + self.MAX_VALID_RESEND_RANGE:
                                message = "Printer request resent for too old line number. Assuming firmware reboot."
                                self.parent.register_error(261, message, is_blocking=True)
                            while self.ready_for_command.acquire(blocking=False):
                              self.logger.info("Flushing command ready semaphore for correct line resend")
                            self.print_thread.line_number_to_resend = failed_line_number
                            self.flush_send_now_buffer()
                            self.logger.info("Due to line resend request switching current line number to %s" % failed_line_number)
                            #self.send_now(b"M110 N%d" % failed_line_number)
                            #while self.ready_for_command.acquire(blocking=False):
                            #    self.logger.info("Flushing command ready semaphore for correct line resend")
                    except AttributeError: # print_thread could be deleted by other thread
                        pass
                else:
                    self.logger.warning("Can't parse line number from resent request: %s" % line)
                return True

    def check_temperature(self, line):
        for prefix in self.TEMPERATURE_ACKS_PREFIXES:
            if prefix in line:
                return self.parse_temperature(line.replace(b'== ', b''))
        return False

    def check_position(self, line):
        if b"X:" in line:
            match = self.position_re.search(line)
            if match:
                self.position = [float(match.group(1)), float(match.group(2)), float(match.group(3)), float(match.group(4))]
                return True
        return False

    def load_gcodes(self, gcodes):
        self.logger.info("Loading gcodes in ThreadedSender")
        with self.thread_start_lock:
            self.stop_temp_requesting()
            time.sleep(self.TEMP_REQUEST_PERIOD)
            if self.print_thread and self.print_thread.is_alive():
                self.logger.info("Joining printing thread...")
                self.print_thread.join(timeout=self.PRINT_JOIN_TIMEOUT)
                if self.print_thread and self.print_thread.is_alive():
                    self.parent.register_error(260, "Error: already printing - unable to start a new print", is_blocking=False)
                    try:
                        send_now_buffer_len = len(self.print_thread.send_now_buffer)
                    except (AttributeError, TypeError):
                        send_now_buffer_len = 0
                    try:
                        lines_sent = self.print_thread.lines_sent
                        total_gcodes = self.print_thread.total_gcodes
                        gcodes_buffer_len = len(self.print_thread.buffer)
                    except (AttributeError, TypeError):
                        lines_sent = 0
                        total_gcodes = 0 
                        gcodes_buffer_len = 0
                    self.logger.warning("Print thread send now: %s and main buffer: %s/%s(%s)" % \
                            (send_now_buffer_len, lines_sent, gcodes_buffer_len, total_gcodes))
                    self.logger.warning("Print thread last line sent: %s" % self.print_thread.last_sent)
                    self.logger.warning("Print thread last line about to be send: %s" % self.print_thread.last_about_to_be_sent)
                    self.logger.warning("Print thread line number to resend:%s" % self.print_thread.line_number_to_resend)
                    self.start_temp_requesting()
                    return False
            self.print_thread = PrintThread(self)
            self.print_thread.load_gcodes(gcodes)
            self.print_thread.start()
            self.init_speed_calculation_thread()
            self.logger.info("Print thread started")
            self.start_temp_requesting()
            return True

    def is_printing(self):
        with self.thread_start_lock:
            return getattr(self, "print_thread", None) and self.print_thread.is_alive() and self.print_thread.buffer

    def is_paused(self):
        with self.thread_start_lock:
            try:
                return self.print_thread and self.print_thread.is_alive() and \
                       (self.print_thread.paused or self.paused_for_user)
            except AttributeError:
                return False

    def is_operational(self):
        return self.operational_flag

    def pause(self, no_gcodes=False):
        if self.print_thread:
            if not self.print_thread.paused:
                if self.heating:
                    message = "Can't pause during heating."
                    self.parent.register_error(254, message, is_blocking=False)
                    return False
                if self.paused_for_user:
                    message = "Can't pause during \"paused for user\"."
                    self.parent.register_error(256, message, is_blocking=False)
                    return False
                self.print_thread.paused = True
                self.was_in_relative_before_pause = self.in_relative_pos_mode
                #self.update_pause_time_and_duration()
                if not no_gcodes:
                    if not self.in_relative_pos_mode:
                        self.send_now(b"G91")
                    self.send_now(self.on_pause_gcodes)
                return True
        return False

    def unpause(self, no_gcodes=False):
        if self.print_thread:
            if self.print_thread.paused:
                if not no_gcodes:
                    resume_gcodes = []
                    for line in self.on_resume_gcodes:
                        line = line.replace(b'{bed_temp}', str(self.target_temps[0]).encode())
                        line = line.replace(b'{t1_temp}', str(self.target_temps[1]).encode())
                        if b"{t2_temp}" in line:
                            if len(self.target_temps) > 2:
                                line = line.replace(b'{t2_temp}', str(self.target_temps[2]).encode())
                            else:
                                continue
                        resume_gcodes.append(line)
                    self.send_now(resume_gcodes)
                    if not self.was_in_relative_before_pause:
                        self.send_now(b"G90")
                self.print_thread.paused = False
                #self.update_unpause_time_and_duration()
                return True
        return False

    def cancel(self):
        if self.profile.get("no_DTR") and self.heating:
            message = "Canceling of heating is not supporting by the printer."
            self.parent.register_error(257, message, is_blocking=True)
            return False
        self.stop_flag = True
        force_board_reset = False
        with self.thread_start_lock:
            if self.print_thread and self.print_thread.is_alive():
                if self.print_thread.line_number_to_resend != None:
                    force_board_reset = True
                self.print_thread.cancel()
                self.print_thread.join(self.STOP_JOIN_TIMEOUT)
                if not force_board_reset: # second time is needed
                    if self.print_thread.line_number_to_resend != None:
                        force_board_reset = True
                self.print_thread = None
        try:
            self.read_thread.join(self.STOP_JOIN_TIMEOUT)
            self.read_thread = None
            self.temperature_request_thread.join()
        except (AttributeError, RuntimeError):
            self.logger.warning("Unable to join read thread")
        try:
            self.temperature_request_thread.join(self.STOP_JOIN_TIMEOUT)
            self.temperature_request_thread = None
        except (AttributeError, RuntimeError):
            self.logger.warning("Unable to join temperature request thread")
        if getattr(self, 'speed_calculation_thread', None):
            self.speed_calculation_thread.join(self.speed_calculation_thread.LOOP_TIME)
        if self.heating or force_board_reset:
            self.reset()
        #TODO try to comment out lines bellow to speed up cancel
        self.stop_flag = False
        self.operational_flag = False
        self.parent.force_cancel_state()
        try:
            self.init()
            self.logger.info("Successful cancel")
        except RuntimeError:
            self.parent.register_error(255, "Can't reconnect to printer after cancel.", is_blocking=True)
        return True

    def reset(self):
        if not self.connection:
            self.logger.warning("No connection to printer to reset")
            return False
        elif self.profile.get('no_DTR'):
            self.logger.warning("DTR reset is not supported for this printer type. Canceling using gcodes.")
            self.connection.send(self.BOARD_RESET_GCODE)
        else:
            self.connection.reset()
            self.logger.info("Successful reset")
        time.sleep(self.RECONNECTION_RESET_SLEEP)

    def get_percent(self):
        if self.print_thread:
            return self.print_thread.get_percent()
        else:
            return 0

    def get_current_line_number(self):
        try:
            lines_sent = self.print_thread.lines_sent
        except AttributeError:
            lines_sent = 0
        return lines_sent

    def unbuffered_gcodes(self, gcodes):
        self.logger.info("Gcodes to send now: " + str(gcodes))
        try:
            gcodes = self.preprocess_gcodes(gcodes)
        except Exception as e:
            self.logger.exception(e)
            return False
        self.logger.info("Gcodes to send now(processed): " + str(gcodes))
        #gcodes.append(self.GET_POSITION_GCODE)
        self.send_now(gcodes)
        self.logger.info("Gcodes were sent to printer")

    #USE only to to fix stupid behaviour as after M28
    def release_ready_for_command(self):
        self.ready_for_command.release()

    def flush_send_now_buffer(self):
        with self.thread_start_lock:
            if self.print_thread and self.print_thread.is_alive():
                try:
                    self.print_thread.send_now_buffer.clear()
                except:
                    pass

    def close(self):
        self.logger.debug("Joining reading thread...")
        self.stop_flag = True
        try:
            self.read_thread.join(6)
        except RuntimeError:
            self.read_thread.join(self.STOP_JOIN_TIMEOUT)
        except (AttributeError, RuntimeError):
            pass
        self.logger.info("...done")
        with self.thread_start_lock:
            if getattr(self, "print_thread", None):
                self.logger.debug("Joining printing thread...")
                try:
                    self.print_thread.join(self.STOP_JOIN_TIMEOUT)
                except (AttributeError, RuntimeError):
                    self.logger.debug("...no running print thread to joint")
                else:
                    self.logger.debug("...done")
        if getattr(self, "temperature_request_thread", None):
            self.logger.debug("Joining temperature thread...")
            try:
                self.temperature_request_thread.join(self.STOP_JOIN_TIMEOUT)
            except (AttributeError, RuntimeError):
                self.logger.debug("...no running temperature thread to join")
            else:
                self.logger.debug("...done")
        if self.connection:
            self.connection.close()
        self.logger.info("Threaded sender is closed")


class PrintThread(threading.Thread):

    RESUME_WAITING_STEP = 0.1
    OK_WAITING_STEP = 0.01
    GO_TO_LINE_N_GCODE = b"M110"
    COMMENT_CHAR = b";"

    #TODO move "with *.thread_start_lock:" here
    def __init__(self, sender, name=None):
        self.sender = sender
        self.lines_sent = 0 # needed for multithreading purposes(lines_sent count be read before load gcodes)
        self.total_gcodes = 0 # needed for multithreading purposes(lines_sent count be read before load gcodes)
        self.paused = False
        self.buffer = None
        self.last_sent = b""
        self.last_about_to_be_sent = b""
        self.send_now_buffer = collections.deque()
        self.line_number_to_resend = None
        self.resend_lock = threading.Lock()
        self.n_offset = 0
        self.skipped_line_numbers = collections.deque()
        if not name:
            name = self.__class__.__name__
        self.logger = sender.logger.getChild(self.__class__.__name__)
        super(PrintThread, self).__init__(name=name, daemon=True)

    def calc_checksum(self, command):
        checksum = 0
        for char in command:
            checksum ^= char
        return checksum

    def load_gcodes(self, gcodes):
        self.logger.info("Starting gcodes loading...")
        self.resend_lock = threading.Lock()
        if type(gcodes) in (bytes, str):
            gcodes = self.sender.preprocess_gcodes(gcodes)
            #self.logger.info('Got %d gcodes to execute.' % len(gcodes))
        self.lines_sent = 0
        gcodes.insert(0, self.GO_TO_LINE_N_GCODE)
        self.total_gcodes = len(gcodes)
        self.sender.total_gcodes = len(gcodes)
        if type(gcodes) == self.sender.BUFFER_CLASS:
            self.buffer = gcodes
        else:
            self.buffer = self.sender.BUFFER_CLASS(gcodes)
        self.n_offset = 0
        self.skipped_line_numbers = collections.deque()
        self.logger.info("...done loading gcodes")

    def add_line_number_and_checksum(self, line):
        # threaded sender if counting all lines(including skipped like comment only or M0/M1)
        # but printer firmware needs get N+1 event if we had skipped a line, so offset is needed
        command = b"N%d %s" % (self.lines_sent-self.n_offset, line)
        command += b"*%d" % self.calc_checksum(command)
        return command

    def strip_comments(self, line):
        return line.split(self.COMMENT_CHAR)[0].strip() 

    def send(self, line):
        self.last_about_to_be_sent = line
        if self.sender.connection.send(line):
            self.sender.analyze_sent_line(line)
            self.last_sent = line
            return True
        self.sender.parent.register_error(250, "Unable to write to serial port", is_blocking=False)
        return False

    def rewind_lines_send_and_offset(self, printers_new_n_line_number):
        while printers_new_n_line_number != self.lines_sent - self.n_offset:
            last_line = self.buffer[self.lines_sent]
            if not self.strip_comments(last_line) or self.check_pause_or_resume(last_line):
                self.n_offset -= 1
            self.lines_sent -= 1
            if self.sender.stop_flag:
                return
            if self.lines_sent == 0:
                self.logger.warning('Reaching a line number zero and resend rewind. Breaking')
                break
            self.logger.debug(f'Rewind lines_sent:{self.lines_sent} offset:{self.n_offset}. Line:{last_line}')
        self.logger.info(f"Resend changed lines_sent:{self.lines_sent} offset:{self.n_offset}")

    @log.log_exception
    def run(self):
        while not self.sender.stop_flag:
            while not self.sender.ready_for_command.acquire(True, self.OK_WAITING_STEP):
                if self.sender.stop_flag:
                    self.sender.ready_for_command.release()
                    return
            if self.send_now_buffer:
                line = self.strip_comments(self.send_now_buffer.popleft())
                if not line or not self.send(line):
                    self.sender.ready_for_command.release()
            elif self.paused or self.sender.paused_for_user:
                time.sleep(self.RESUME_WAITING_STEP)
                self.sender.ready_for_command.release()
            else:
                try:
                    with self.resend_lock:
                        if self.line_number_to_resend != None:
                            self.rewind_lines_send_and_offset(self.line_number_to_resend)
                            self.line_number_to_resend = None
                    if not self.buffer or self.total_gcodes == self.lines_sent:
                        raise IndexError #last gcode line is sent - finish the print
                    line = self.buffer[self.lines_sent]
                    if self.sender.heating_emulator:
                        line = self.sender.heating_emulator.start(line)
                except (IndexError, TypeError):
                    #self.logger.debug(threading.currentThread().getName() + " has finished.")
                    if self.buffer:
                        self.logger.info(f"Print finished on %d/%d." % (self.lines_sent, self.total_gcodes))
                        # self.logger.debug("Print finished on %d/%d." % (self.lines_sent, self.total_gcodes))
                    self.buffer = None
                    self.sender.ready_for_command.release()
                    return
                else:
                    str_line = self.strip_comments(line)
                    if str_line and not self.check_pause_or_resume(str_line):
                        line = self.add_line_number_and_checksum(str_line)
                        if self.send(line):
                            self.lines_sent += 1
                            continue
                    else:
                        self.n_offset += 1
                        self.logger.debug(f'Skipped lines offset: {self.n_offset}. Line: {line}')
                    self.lines_sent += 1
                    self.sender.ready_for_command.release()

    def pause(self):
        self.paused = True

    def unpause(self):
        self.paused = False

    def cancel(self):
        try:
            self.buffer.clear()
        except:
            pass
        try:
            self.send_now_buffer.clear()
        except:
            pass

    def get_percent(self):
        if not self.lines_sent or\
           not self.total_gcodes or\
           self.lines_sent < 0:
            return 0
        else:
            return round(self.lines_sent / self.total_gcodes * 100, 2)

    def check_pause_or_resume(self, line):
        if self.sender.intercept_pause:
            for gcode in self.sender.pause_gcodes_list:
                if self.line_contains_gcode(line, gcode) and not self.ignore_pause_resume(line):
                    self.sender.parent.register_error(601, f"Pause by gcode {line}", is_blocking=False, is_info=True)
                    self.sender.pause()
                    return True
            for gcode in self.sender.resume_gcodes_list:
                if self.line_contains_gcode(line, gcode) and not self.ignore_pause_resume(line):
                    self.sender.parent.register_error(602, f"Resume by gcode {line}", is_blocking=False, is_info=True)
                    self.sender.unpause()
                    return True
        return False

    def line_contains_gcode(self, line, gcode):
        return line.endswith(gcode) or\
            gcode + b";" in line or\
            gcode + b" " in line 

    def ignore_pause_resume(self, line):
        return self.sender.IGNORE_PAUSE_MAGIC in line or\
               bool(self.sender.ignore_pause_resume_re.search(line))
